[<Microsoft.FSharp.Core.AutoOpen>]
module Degenz.Types

open System

[<Measure>]
type mins

[<Measure>]
type GBT

type Range = { Min : int ; Max : int }

module Range =
    let normalized = { Min = 0 ; Max = 100 }
    let create min max = { Min = min ; Max = max }
    let constrain value range = if value < range.Min then range.Min elif value > range.Max then range.Max else value

type ItemId =
    | Virus = 0
    | RemoteAccess = 1
    | Worm = 2
    | Firewall = 6
    | Encryption = 7
    | Cypher = 8
    | ProteinPowder = 12
    | ToroLoco = 13
    | Cigs = 14
    | MoonPie = 15

type StatId =
    | Strength = 0
    | Focus = 1
    | Charisma = 2
    | Luck = 3

type StatConfig = {
    Id : StatId
    BaseDecayRate : float
    BaseRange : Range
}

type PlayerStat = {
    Id : StatId
    Amount : int
    ModRange : Range
    LastRead : DateTime
}
with static member empty = { Id = StatId.Strength ; Amount = 0 ; ModRange = Range.normalized ; LastRead = DateTime.UtcNow}

type Stats = {
    Strength : PlayerStat
    Focus : PlayerStat
    Luck : PlayerStat
    Charisma : PlayerStat
}
with static member empty = { Strength = PlayerStat.empty ; Focus = PlayerStat.empty ; Luck = PlayerStat.empty ; Charisma = PlayerStat.empty }

type HackResult =
    | Strong
    | Weak

type DiscordPlayer = { Id: uint64; Name: string }
    with static member empty = { Id = 0uL ; Name = "None" }

type HackEvent = {
    IsInstigator : bool
    Adversary : DiscordPlayer
    Success : bool
    HackId : int
}

type PlayerEventType =
    | Hacking of HackEvent
    | Shielding of shieldId : int
    | Stealing of instigator : bool * adversary : DiscordPlayer
    | PlayingSlot
    | Imprison

type PlayerEvent =
    { Type : PlayerEventType
      Cooldown : int<mins>
      Timestamp : DateTime }

[<RequireQualifiedAccess>]
type ItemType =
    | Hack
    | Shield
    | Food
    | Accessory
    | Jpeg
    | Misc
    
type Effect =
    | Min of int
    | Max of int
    | Add of int
    | RateMultiplier of float
    
type StatEffect = {
    TargetStat : StatId
    Effect : Effect
}

type ItemAttribute =
    | Buyable of price : int<GBT>
    | Sellable of price : int<GBT>
    | RateLimitable of cooldown : int<mins>
    | Expireable of lifetime : int<mins>
    | Consumable
    | Modifiable of effects : StatEffect list
    | Droppable of chance : float
    | Tradeable
    | Attackable of power : int
    | Defendable of resistance : int
    | Classable of className : string
    | Stackable of max : int

let (|CanBuy|_|) itemAttrs       = itemAttrs |> List.tryPick (function Buyable       p  -> Some p  | _ -> None)
let (|CanSell|_|) itemAttrs      = itemAttrs |> List.tryPick (function Sellable      p  -> Some p  | _ -> None)
let (|CanExpire|_|) itemAttrs    = itemAttrs |> List.tryPick (function Expireable    l  -> Some l  | _ -> None)
let (|CanRateLimit|_|) itemAttrs = itemAttrs |> List.tryPick (function RateLimitable l  -> Some l  | _ -> None)
let (|CanConsume|_|) itemAttrs   = itemAttrs |> List.tryPick (function Consumable       -> Some () | _ -> None)
let (|CanModify|_|) itemAttrs    = itemAttrs |> List.tryPick (function Modifiable    es -> Some es | _ -> None)
let (|CanDrop|_|) itemAttrs      = itemAttrs |> List.tryPick (function Droppable     c  -> Some c  | _ -> None)
let (|CanTrade|_|) itemAttrs     = itemAttrs |> List.tryPick (function Tradeable        -> Some () | _ -> None)
let (|CanAttack|_|) itemAttrs    = itemAttrs |> List.tryPick (function Attackable    p  -> Some p  | _ -> None)
let (|CanDefend|_|) itemAttrs    = itemAttrs |> List.tryPick (function Defendable    r  -> Some r  | _ -> None)
let (|CanClass|_|) itemAttrs     = itemAttrs |> List.tryPick (function Classable     c  -> Some c  | _ -> None)
let (|CanStack|_|) itemAttrs     = itemAttrs |> List.tryPick (function Stackable     m  -> Some m  | _ -> None)
    
type Item = {
    Id : int
    Name : string
    Description : string
    Type : ItemType
    Symbol : string
    IconUrl : string
    Attributes : ItemAttribute list
}

type StoreItem = {
    Stock : int
    LimitStock : bool
    Available : bool
    Item : Item
}

type HackItem = {
    Id : int
    Name : string
    Price : int<GBT>
    Cooldown : int<mins>
    Power : int
    Class : string
}

type ShieldItem = {
    Id : int
    Name : string
    Price : int<GBT>
    Cooldown : int<mins>
    Resistance : int
    Class : string
}

type Inventory = Item list

type PlayerData = {
    DiscordId : uint64
    Name : string
    Inventory : Inventory
    Events : PlayerEvent list
    Stats : Stats
    Bank : int<GBT>
    Active : bool
}
//        Achievements : string array
//        XP : int
with member this.toDiscordPlayer = { Id = this.DiscordId ; Name = this.Name }
     static member empty =
         { DiscordId = 0uL
           Name = "None"
           Inventory = []
           Events = []
           Stats = Stats.empty
//               Achievements = [||]
//               XP = 0
           Bank = 0<GBT>
           Active = false }

